(function scriptedWebServiceOperation(request, response) {

	response.log_messages="--Check Import WS Started--\n";
	response.log_messages+="   WS Name: " + request.import_ws_name + "\n";
	response.log_messages+="   Target Table: " + request.table_name + "\n";
	response.log_messages+="   Scope: " + request.scope + "\n";
	response.log_messages+="   Run Business Rules: " + request.run_business_rules + "\n";
	try {
		//validation
		//validate scope
		var grApplication=new GlideRecord("sys_scope");
		grApplication.addEncodedQuery("name="+request.scope);
		grApplication.query();
		if(!grApplication.next()) {
			response.status="ERROR";
			response.status_msg="scope parameter: Application " + request.scope + "does not exist on this server";
			return;
		}
		var scope=grApplication.sys_id;
		//validate table name
		var grTable=new GlideRecord("sys_db_object");
		var import_ws_exists=false;
		var inc_business_rule=false;
		var table_label;
		var tableList;
		grTable.addEncodedQuery("name="+request.table_name);
		grTable.query();
		if(!grTable.hasNext()) {
			response.status="ERROR";
			response.status_msg="table_name parameter: Table " + request.table_name + "does not exist on this server";
			return;
		}
		else {
			grTable.next();
			table_label=grTable.label;
			tableList=getTargetTableList(request.table_name,response);
		}
		//check the value of include business rules flag
		inc_business_rule=convertBoolean(request.run_business_rules);
		if(inc_business_rule==null) {
			response.status="ERROR";
			response.status_msg="include_business_rules parameter: Invalid value";
			return;
		}
		//check for existing import set ws
		grTable.addEncodedQuery("name="+request.import_ws_name);
		grTable.query();
		if(grTable.hasNext()) {
			import_ws_exists=true;
		}
		//create the import set ws
		if(!import_ws_exists) {
			createImportSetWS(request.import_ws_name,request.table_name,table_label,scope, response);
		}
		//synchronise the fields on the ws with those on the source table
		createWSFields(request.import_ws_name,tableList,scope,response);
		//check/create transform map
		createTransformMap(request.import_ws_name,request.table_name,request.run_business_rules,response,scope);
		response.log_messages+="--Check Import Set WS and Transform Map Completed--\n";
		response.status="SUCCESS";
	}
	catch(e) {
		response.status="ERROR";
		response.status_msg=response.status_msg+"\n"+ e.message;
	}

})(request, response);

function createImportSetWS (importSetWSName,targetTableName,targetTableLabel,scope,response) {
	//create the import set ws
	var grCreateWS=new GlideRecord("v_ws_creator");
	grCreateWS.initialize();
	grCreateWS.name=importSetWSName
	grCreateWS.label="ITSM Bridge: "+targetTableLabel;
	grCreateWS.sys_scope=scope;
	grCreateWS.copy_fields="true";
	grCreateWS.create_transform_map="true";
	grCreateWS.extends_table="sys_import_set_row";
	grCreateWS.target_table=targetTableName;
	grCreateWS.insert();
	response.log_messages+="--created ws import set: " + importSetWSName+" ["+ targetTableName+"]--\n"
}

function addFieldToAttributeList(attrs,fieldName, fieldLength) {
	var ca=new GlideColumnAttributes(fieldName);
	ca.setType("string");
	ca.setUsePrefix(false);
	ca.setLength(fieldLength);
	attrs.put(fieldName, ca);
}

function createTransformMap(sourceTable, targetTable, includeBusinessRules,response, scope){
	//check/create transform map
		var grTransformMap=new GlideRecord("sys_transform_map");
		grTransformMap.addEncodedQuery("name="+sourceTable);
		grTransformMap.addEncodedQuery("source_table="+sourceTable);
		grTransformMap.addEncodedQuery("active=true");
		grTransformMap.query();
		if(!grTransformMap.next()) {
			grTransformMap.initialize();
		}
		grTransformMap.name=sourceTable;
		grTransformMap.sys_name=sourceTable;
		grTransformMap.source_table=sourceTable;
        grTransformMap.sys_scope=scope;
		grTransformMap.target_table=targetTable;
		grTransformMap.sys_customer_update=true;
		grTransformMap.run_business_rules=includeBusinessRules;
		grTransformMap.copy_empty_fields=false;
		grTransformMap.enforce_mandatory_fields="All Fields";
		if(grTransformMap.run_script==false) {
			if(grTransformMap.isNewRecord()) {
				grTransformMap.insert();
				response.log_messages+="--transform map created: " + sourceTable +"\n";
			}
			else {
				grTransformMap.update();
				response.log_messages+="--transform map updated: " + sourceTable +"\n";
			}
		}
		else {
			response.log_messages+="--transform is scripted - no changes made--\n";
			return; //do not update or add fields if script is present
		}
		var map_sys_id=grTransformMap.sys_id;
		//check/create transform map mappings
		var sourceGR = new GlideRecord(sourceTable);
		sourceGR.initialize();
		var targetGR = new GlideRecord(targetTable);
		targetGR.initialize();  
		var targetGRElements = targetGR.getFields();
		var fields = sourceGR.getFields();
		for (var i = 0; i < fields.size(); i++) {
			var field = fields.get(i);
			var name = field.getName();
			var label = field.getED().getLabel();
			//skip virtual fields like Tags
			if (!field.getED().isInDatabase()) {
				response.log_messages+="   skip virtual field: " + name+"\n";
				continue;
			}
			var tname = matchFieldName(name, targetGRElements);
			
			if (!isSkip(name) && tname != null && !isSkip(tname)) {
				createTransformEntry(name, tname,sourceTable,targetTable,map_sys_id,false,response,scope);
			}
			else {
				response.log_messages+="   skip system field: " + name+"\n";
			}
		}
	//create additional transform map entries for system fields (if required)
	if(hasField("sys_scope",targetGRElements)) 
		createTransformEntry("target_scope", "sys_scope",sourceTable,targetTable,map_sys_id,false,response,scope);
	if(hasField("sys_domain",targetGRElements)) 
		createTransformEntry("target_domain", "sys_domain",sourceTable,targetTable,map_sys_id,false,response,scope);
	if(hasField("sys_domin_path",targetGRElements)) 
		createTransformEntry("target_domain_path", "sys_domain_path",sourceTable,targetTable,map_sys_id,false,response,scope);
	if(hasField("sys_domin_path",targetGRElements)) 
		createTransformEntry("target_domain_path", "sys_domain_path",sourceTable,targetTable,map_sys_id,false,response,scope);
	if(hasField("sys_class_name",targetGRElements)) 
		createTransformEntry("target_class_name", "sys_class_name",sourceTable,targetTable,map_sys_id,false,response,scope);
	createTransformEntry("target_mod_count", "sys_mod_count",sourceTable,targetTable,map_sys_id,false,response,scope);
	createTransformEntry("target_created_by", "sys_created_by",sourceTable,targetTable,map_sys_id,false,response,scope);
	createTransformEntry("target_created_on", "sys_created_on",sourceTable,targetTable,map_sys_id,false,response,scope);
	if(!hasExistingCoalece(map_sys_id)) {
		response.log_messages+="add sys_id_coalece\n";
		createTransformEntry("target_sys_id", "sys_id",sourceTable,targetTable,map_sys_id,true,response,scope);
	}
	else {
		response.log_messages+="coalece field exists\n";
	}
	response.log_messages+="--synchronise transform map complete--\n";
}

//construct the list of field source tables by following the table extensions back
function getTargetTableList(tableName,response) {
	var grTable=new GlideRecord("sys_db_object");
	grTable.addEncodedQuery("name="+tableName);
	grTable.query();
	if(grTable.next()) {
		var tableExtends=grTable.super_class.name;
		var tableList=grTable.name;
		while(grTable.super_class.name!=""){
			tableList=tableList+","+tableExtends;
			grTable=new GlideRecord("sys_db_object");
			grTable.addEncodedQuery("name="+tableExtends);
			grTable.query();
			if(grTable.next()) {
				tableExtends=grTable.super_class.name;
			}
		}
		response.log_messages+="   include fields from tables: " + tableList +"\n";
		return tableList;
	}
	else {
		return null;
	}
}

function createWSFields(wsTable,targetTableList,scopeID,response) {
	var grDict=new GlideRecord("sys_dictionary");
	var attrs = new Packages.java.util.HashMap();
	grDict.addEncodedQuery("nameIN"+targetTableList+"^internal_type!=collection");
	grDict.query();
	var ca;
	while(grDict.next())
	{
		if(!isSkip(grDict.element)){
			ca = new GlideColumnAttributes(grDict.element);
			ca.setType("string");
			ca.setUsePrefix(false);
			if(grDict.max_length!=null) {
				ca.setLength(grDict.max_length);
			}
			else {
				ca.setLength("200");
			}
			attrs.put(grDict.element, ca);
		}
	}
	//add a scope field
	addFieldToAttributeList(attrs,"target_scope","100");
	//add a created by field
	addFieldToAttributeList(attrs,"target_created_by","100");
	//add a created on field
	addFieldToAttributeList(attrs,"target_created_on","100");
	//add a target_sys_id field
	addFieldToAttributeList(attrs,"target_sys_id","100");
	//add a domain field
	addFieldToAttributeList(attrs,"target_domain","100");
	//add a domain  path field
	addFieldToAttributeList(attrs,"target_domain_path","100");
	//add a class_name field
	addFieldToAttributeList(attrs,"target_class_name","100");
	//add a mod_count field
	addFieldToAttributeList(attrs,"target_mod_count","100");
        //add source identifier
        addFieldToAttributeList(attrs,"source_id","100");
        //add batch identifier
        addFieldToAttributeList(attrs,"batch_id","100");
	var tc = new GlideTableCreator(wsTable,wsTable);
	tc.setColumnAttributes(attrs);
	tc.setExtends("sys_import_set_row");
	tc.update();
	
	//now change any incorrect scope values
	var grDictionary=new GlideRecord('sys_dictionary');
	grDictionary.addEncodedQuery("name="+wsTable+"^sys_scope!="+scopeID);
	grDictionary.query();
	while(grDictionary.next()) {
		grDictionary.sys_scope=scopeID;
		grDictionary.update();
		response.log_messages+="updated scope: " + grDictionary.element+"["+scopeID+"]\n";
	}
	response.log_messages+="--synchronise ws fields with target complete--\n";
}

function isSkip(name) {
    // skip sys fields
    if (name == "sys_created_on" ||
        name == "sys_updated_on" ||
        name == "sys_created_by" ||
        name == "sys_updated_by" ||
        name == "sys_mod_count" ||
        name == "sys_class_name" ||
		name == "sys_scope" ||
		name == "sys_domain" ||
		name == "sys_domain_path" ||
        name == "sys_id" ||
        name == "sys_import_set" ||
        name == "sys_import_row" ||
        name == "sys_import_state")
      return true;

    return false;
}

function matchFieldName(name, targetGRElements) {
  for(var i = 0; i < targetGRElements.size(); i++) {
	 var field = targetGRElements.get(i);
     if (name == field.getName()) {
       return field.getName();
     }
  }
  return null;
}

function hasField(name, fieldList) {
  for(var i = 0; i < fieldList.size(); i++) {
	 var field = fieldList.get(i);
     if (name == field.getName()) {
       return true
     }
  }
  return false;
}

function convertBoolean(strBoolean) {
	switch(strBoolean.toUpperCase()) {
		case "YES":
		case "TRUE":
		case "1":
			return true;
		case "NO":
		case "FALSE":
		case "0":
			return false;
		default:
		    return null;
	}
}

function createTransformEntry(sourceName, targetName,source_table, target_table, map_sys_id,coalesce,response, scope) {
  var entryGR = new GlideRecord("sys_transform_entry");
  entryGR.addQuery("map",map_sys_id);
  entryGR.addQuery("source_table",source_table);
  entryGR.addQuery("source_field",sourceName);
  entryGR.addQuery("target_table",target_table);
  entryGR.addQuery("target_field",targetName);
  entryGR.query();
  if (!entryGR.next()) {
	response.log_messages+="   create transform field mapping: " + sourceName+" -> "+ targetName+"\n";
    entryGR.map = map_sys_id;
    entryGR.source_table = source_table;
    entryGR.source_field = sourceName;
    entryGR.target_table = target_table;
    entryGR.target_field = targetName;
    entryGR.choice_action = 'ignore';
    entryGR.coalesce = coalesce;
    entryGR.sys_scope = scope;
    entryGR.insert();
  }
	else {
		response.log_messages+="   transform field mapping exists: " + sourceName+" -> "+ targetName+"\n";
	}
}

function hasExistingCoalece(map_sys_id) {
    var entryCoalece = new GlideRecord("sys_transform_entry");
    entryCoalece.addEncodedQuery("map="+map_sys_id+"^coalesce=true");
	entryCoalece.query();
    return entryCoalece.next();
    
}