/* 
 * Copyright (c) Service Management Integrations Limited, 2014 - 2018
 * All Rights Reserved.
 * NOTICE: All information contained herein is and remains the property of Service Management  Integrations Limited.  The intellectual and technical concepts contained herein are proprietary to Service Management Integrations Limited and are protected by copyright law.  Dissemination of this information or reproduction of this material is strictly forbidden unless prior written permission is obtained from Service Management Integrations Limited.
 */

var setHRTemplateDependencies = Class.create();
setHRTemplateDependencies.prototype = {
    initialize: function() {
		this.logEntries=String('');
		this.status='ok';
    },
	
	type: 'setHRTemplateDependencies',
	
	setHRTemplateTaskDependencies: function (sourceId,targetId) {
		try {
			this.log('processing: [src='+sourceId+' tgt='+targetId+']');
			var sourceModelName=this.getHRCaseModelName(sourceId);
			var taskData=this.getSourceTasks(sourceId);
			this.addDependencies(sourceId,taskData);
			this.addCaseDependencies(taskData);
			this.setNextTaskValues(taskData);
			this.setNextChildTaskValues(taskData);
			this.addTargetTaskIDs(taskData,sourceModelName,targetId);
			this.setTargetNextTaskReferences(taskData);
			this.log('processing completed');
		}
		catch(e) {
			this.error(e.message);
		}
	},
	
	getHRCaseModelName: function(sourceId) {
		var gr=new GlideRecord('cmdb_hr_case_product_model');
		gr.get(sourceId);
		this.log('Source Case Model: '+gr.name);
		return String(gr.name);
	},
	
	//retrieves an array of tasks that are defined for the source template.
	getSourceTasks : function(sourceID) {
		var gr = new GlideRecord('sm_m2m_somodel_stmodel');
		gr.addEncodedQuery('so_model='+sourceID);
		gr.orderBy('order');
		gr.orderByDesc();
		gr.query();
		var taskCount=gr.getRowCount();
		var ta = new Array(taskCount+1);
		ta[0]={taskID:sourceID,name:'CASE',order:-1};
		ta[0].upstreamTasks={};
		ta[0].downstreamTasks={};
		for(var i=1;i<taskCount+1;i++) {
			gr.next();
			ta[i]={taskID:String(gr.st_model),name:String(gr.st_model.name), order: Number(gr.order)};
			ta[i].upstreamTasks={};
			ta[i].downstreamTasks={};
		}
		this.log('Source task count: '+ta.length);
		return ta;
	},
	//retrieves an array of task dependencies that are defined for the source  template.
	//then appends the upstream ands downstream task values to each of the tasks
	addDependencies: function(sourceID,sourceTasks) {
		var gr = new GlideRecord('sm_m2m_task_template_dependency');
		gr.addQuery('order_model',sourceID);
		gr.query();
		var depCount=gr.getRowCount();
		var da=new Array(depCount);
		for(var i=0;i<depCount;i++) {
			gr.next();
			var upstreamTask=this.getTask(sourceTasks,gr.upstream_task);
			var downstreamTask=this.getTask(sourceTasks,gr.downstream_task);
			if(upstreamTask!=undefined) {
				upstreamTask.downstreamTasks[gr.downstream_task]=downstreamTask;
			}
			else {
				this.warning('no matching upstream task for id='+gr.upstream_task);
			}
			if(downstreamTask!=undefined) {
				downstreamTask.upstreamTasks[gr.upstream_task]=upstreamTask;
			}
			else {
				this.warning('no matching downstream task for id='+gr.downstream_task);
			}
		
		}
		this.log('task dependencies collected and processed ['+da.length+']');
	},
	//all source tasks that do not have upstream dependencies are depenent on the case
	addCaseDependencies: function(sourceTasks) {
		for(var i=1;i<sourceTasks.length;i++) {
			if(Object.keys(sourceTasks[i].upstreamTasks).length==0) {
				sourceTasks[0].downstreamTasks[sourceTasks[i].taskID]=sourceTasks[i];
				sourceTasks[i].upstreamTasks[sourceTasks[0].taskID]=sourceTasks[0];
			}
		}
		this.log('Case dependencies deduced ['+Object.keys(sourceTasks[0].downstreamTasks).length+']');
	},

    getTask: function(sourceTasks,taskID) {
		for(var i=0;i<sourceTasks.length;i++) {
			if(sourceTasks[i].taskID==taskID) {
				return sourceTasks[i];
			}
		}
		return null;
	},
	
	//where a task has more than one downstream tasks, these are linked together by setting the 'next'value
	setNextTaskValues: function(taskData) {
		for(var i=0;i<taskData.length;i++) {
			var dCount=Object.keys(taskData[i].downstreamTasks).length;
			if(dCount>1) {
				var downstreamTaskArray=this.getOrderedTaskArray(taskData[i].downstreamTasks);
				//downstreamTaskArray.sort(this.taskCompare);
				for(var j=0;j<downstreamTaskArray.length-1;j++) {
					downstreamTaskArray[j].next=downstreamTaskArray[j+1];
				}
			}
		}
		this.log('Next Task values calculated');
	},
	
	setNextChildTaskValues: function(taskData) {
		for(var i=0;i<taskData.length;i++) {
			var upstreamTaskArray=this.getOrderedTaskArray(taskData[i].upstreamTasks);
			for(var j=0;j<upstreamTaskArray.length;j++) {
				if(upstreamTaskArray[j].nextChild==undefined) {
					upstreamTaskArray[j].nextChild=taskData[i];
				}
			}
		}
		this.log('Next Child Task values calculated');
	},
	
	taskCompare : function(task1,task2) {
		return task1.order-task2.order;// asc order
	},
	
	getOrderedTaskArray: function (taskObject) {
		var keyArray=Object.keys(taskObject);
		var valueArray=new Array(keyArray.length);
		for(var i=0;i<valueArray.length;i++) {
			valueArray[i]=taskObject[keyArray[i]];
		}
		return valueArray.sort(this.taskCompare);
	},
	
	//match the taskArrayEntries with the corresponding target records and set the targetID
	addTargetTaskIDs: function(taskArray,sourceTemplateName,targetCaseId) {
		
		for(var i=0;i<taskArray.length;i++) {
			
			if(i==0) { //case
				taskArray[i].targetID=targetCaseId;
			}
			else { //tasks
				var querystr='name='+taskArray[i].name.substring(0,60);
				querystr+='^table=sn_hr_core_task^short_description='+sourceTemplateName.substring(0,49)+'->'+taskArray[i].name.substring(0,49);
				var gr=new GlideRecord('sn_hr_core_template');
				gr.setWorkflow(false);
				gr.addEncodedQuery(querystr);
				gr.query();
				if(gr._next()) {
					taskArray[i].targetID=String(gr.sys_id);
				}
				else {
					this.warning('No matching target found for task :'+taskArray[i].name);
				}
			}
		}
		this.log('Matched source tasks with Target records');
	},
	
	//for each next and nextChild value,  
	setTargetNextTaskReferences: function(taskArray) {
		for(var i=0;i<taskArray.length;i++) {
			var gr=new GlideRecord('sn_hr_core_template');
			gr.setWorkflow(false);
			gr.get(taskArray[i].targetID);
			if(gr.sys_id) {
				if(taskArray[i].next!=undefined) {
					gr.next=taskArray[i].next.targetID;
				}
				else {
					gr.next='NULL';
				}
				if(taskArray[i].nextChild!=undefined) {
					gr.next_child=taskArray[i].nextChild.targetID;
				}
				else {
					gr.next_child='NULL';
				}
				gr.update();
			}
		}
		this.log('Set next and next_child Task references on case and task templates completed');
	},
	
	log: function(logEntry) {
		this.logEntries+=logEntry+'\n';
	},
	warning : function(msg) {
		this.log('WARNING: '+msg);
		this.status='warning';
	},
	error : function(msg) {
		this.log('ERROR:'+msg);
		this.status='error';
	},

	getStatus: function() {
		return this.status;
	},

	getLog:  function() {
		return this.logEntries;
	}
	
};


